--[[ TODO:
	Add class default profiles
	Fix no stats in recipe tooltips
	Fix weapon type localization
	Don't show comparison for second ring/trinket/weapon when hovering over first equipped one
	Fix problems with a few random-enchant weapons saying they have 10000+ stats
	
CHANGELOG:
		1.0	- Initial release
		1.1 - Added Attack Power, PvP Power/Resilience, Resistances, and Spell Penetration, added more support for random suffixes
]]

SimpleStats = LibStub("AceAddon-3.0"):NewAddon("SimpleStats", "AceConsole-3.0", "AceEvent-3.0")
local handled, addedLine, prettyName, curStats, newStats, invTypes, tooltip, order

local defaults = {
	profile = {
		armor = true,
		dps = true,
		metasockets = true,
		prismaticsockets = true,
		cogwheelsockets = true,
		redsockets = true,
		bluesockets = true,
		yellowsockets = true,
		stamina = true,
		agility = true,
		strength = true,
		intellect = true,
		spirit = true,
		hit = true,
		crit = true,
		haste = true,
		expertise = true,
		dodge = true,
		parry = true,
		mastery = true,
		spellpower = true,
		attackpower = true,
		pvppower = true,
		pvpresilience = true,
		resistance = true,
		spellpenetration = true,
		
		cloth = true,
		leather = true,
		mail = true,
		plate = true,
		shields = true,
		
		usableweaponsonly = true,
		minlevel = false,
		minquality = 1
	}
}

local options = {
	name = "SimpleStats",
	handler = SimpleStats,
	type = "group",
	args = {		
		minquality = {
			type = "select",
			name = "Minimum Quality",
			values = {"Uncommon", "Rare", "Epic"},
			get = "get",
			set = "set",
			order = 2,
		},
		
		usableweaponsonly = {
			type = "toggle",
			name = "Ignore unusable weapon types",
			get = "get",
			set = "set",
			order = 2.5,
			width = "full"
		},
		
		header2 = {
			type = "header",
			name = "Armor Types",
			order = 3
		},
		cloth = {
			type = "toggle",
			name = "Show on cloth",
			get = "get",
			set = "set",
			order = 4,
		},
		leather = {
			type = "toggle",
			name = "Show on leather",
			get = "get",
			set = "set",
			order = 5,
		},
		mail = {
			type = "toggle",
			name = "Show on mail",
			get = "get",
			set = "set",
			order = 6,
		},
		plate = {
			type = "toggle",
			name = "Show on plate",
			get = "get",
			set = "set",
			order = 7,
		},
		shields = {
			type = "toggle",
			name = "Show on shields",
			get = "get",
			set = "set",
			order = 7.5,
		},
		
		header3 = {
			type = "header",
			name = "Stats",
			order = 8
		},
		
		armor = {
			type = "toggle",
			name = "Armor",
			get = "get",
			set = "set",
			order = 9,
			width = "half"
		},
		dps = {
			type = "toggle",
			name = "DPS",
			get = "get",
			set = "set",
			order = 9.5,
			width = "half"
		},
		stamina = {
			type = "toggle",
			name = "Stamina",
			get = "get",
			set = "set",
			order = 10,
			width = "half"
		},
		agility = {
			type = "toggle",
			name = "Agility",
			get = "get",
			set = "set",
			order = 11,
			width = "half"
		},
		strength = {
			type = "toggle",
			name = "Strength",
			get = "get",
			set = "set",
			order = 12,
			width = "half"
		},
		intellect = {
			type = "toggle",
			name = "Intellect",
			get = "get",
			set = "set",
			order = 13,
			width = "half"
		},
		
		d1={type="description",name=" ",order=13.5},
		
		metasockets = {
			type = "toggle",
			name = "Meta Sockets",
			get = "get",
			set = "set",
			order = 14,
		},
		prismaticsockets = {
			type = "toggle",
			name = "Prismatic Sockets",
			get = "get",
			set = "set",
			order = 15,
		},
		cogwheelsockets = {
			type = "toggle",
			name = "Cogwheel Sockets",
			get = "get",
			set = "set",
			order = 16,
		},
		redsockets = {
			type = "toggle",
			name = "Red Sockets",
			get = "get",
			set = "set",
			order = 17,
		},
		bluesockets = {
			type = "toggle",
			name = "Blue Sockets",
			get = "get",
			set = "set",
			order = 18,
		},
		yellowsockets = {
			type = "toggle",
			name = "Yellow Sockets",
			get = "get",
			set = "set",
			order = 19,
		},
		
		d2={type="description",name=" ",order=20},
		
		hit = {
			type = "toggle",
			name = "Hit",
			get = "get",
			set = "set",
			order = 21,
		},
		crit = {
			type = "toggle",
			name = "Crit",
			get = "get",
			set = "set",
			order = 22,
		},
		haste = {
			type = "toggle",
			name = "Haste",
			get = "get",
			set = "set",
			order = 23,
		},
		spirit = {
			type = "toggle",
			name = "Spirit",
			get = "get",
			set = "set",
			order = 24,
		},
		spellpower = {
			type = "toggle",
			name = "Spell Power",
			get = "get",
			set = "set",
			order = 25,
		},
		mastery = {
			type = "toggle",
			name = "Mastery",
			get = "get",
			set = "set",
			order = 26,
		},
		expertise = {
			type = "toggle",
			name = "Expertise",
			get = "get",
			set = "set",
			order = 27,
		},
		dodge = {
			type = "toggle",
			name = "Dodge",
			get = "get",
			set = "set",
			order = 28,
		},
		parry = {
			type = "toggle",
			name = "Parry",
			get = "get",
			set = "set",
			order = 29,
		},
		
		attackpower = {
			type = "toggle",
			name = "Attack Power",
			get = "get",
			set = "set",
			order = 30,
		},
		
		pvppower = {
			type = "toggle",
			name = "PvP Power",
			get = "get",
			set = "set",
			order = 31,
		},
		
		pvpresilience = {
			type = "toggle",
			name = "PvP Resilience",
			get = "get",
			set = "set",
			order = 32,
		},
		
		resistance = {
			type = "toggle",
			name = "Resistances",
			get = "get",
			set = "set",
			order = 33,
		},
		
		spellpenetration = {
			type = "toggle",
			name = "Spell Penetration",
			get = "get",
			set = "set",
			order = 34,
		},
		
	}
}

function SimpleStats:get(i)
	return self.db.profile[i[1]]
end
function SimpleStats:set(i,v)
	self.db.profile[i[1]] = v
end

function SimpleStats:ChatCommand(input)
	input = input:trim()
	if not input or input == "" then
		InterfaceOptionsFrame_OpenToCategory(SimpleStats.optionsFrame)
	
	elseif (input == "v" or input == "input" or input == "input") then
		self:Print("Version: "..GetAddOnMetadata("SimpleStats","Version"))
	
	elseif (input == "profile" or input == "profiles") then
		InterfaceOptionsFrame_OpenToCategory(SimpleStats.profileFrame)
	
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(SimpleStats, "ss", "SimpleStats", input)
	end
end

local function sortStats(t)
	local tp = {} -- positives
	local tn = {} -- negatives
	local tf = {} -- final
	
	-- Sort into two piles, positives and negatives
	for k,v in pairs(t) do
		v = v+0
		if v > 0 then
			table.insert(tp,{k,v})
		elseif v < 0 then
			table.insert(tn,{k,v})
		end
	end
	
	-- Now sort each pile by stat name
	for k,v in pairs(tp) do
		local name = v[1]
		local value = math.floor(v[2])
		if order[prettyName[name]] then
			local pos = order[prettyName[name]]
			tf[pos] = {name,value}
		end
	end
	
	for k,v in pairs(tn) do
		local name = v[1]
		local value = math.floor(v[2])
		if order[prettyName[name]] then
			local pos = order[prettyName[name]]+30
			tf[pos] = {name,value}
		end
	end
	
	return tf
end

local function resetTooltip()
	handled = false
	addedLine = false
end

local function statIsEnabled(name)
	if strmatch(name,"Resistance") then
		return SimpleStats.db.profile.resistance
	else
		return SimpleStats.db.profile[name:lower():gsub(" ","")]
	end
end

local function printStats(tnew,tcur,tcur2)
	local tchanged = {}
	
	if tcur2 then -- Comparing two existing items, so merge tcur2 into tcur
		for k,v in pairs(tcur2) do
			if tcur[k] then
				tcur[k] = tcur[k] + v
			else
				tcur[k] = v
			end
		end
	end

	for k,v in pairs(tnew) do
		if tcur[k] then -- Stat exists on both, do subtraction
			tchanged[k] = tnew[k] - tcur[k]
		else -- New stat, just use the value
			tchanged[k] = tnew[k]
		end
	end
	
	for k,v in pairs(tcur) do
		if not tnew[k] then
			tchanged[k] = -v
		end
	end
	
	local stats = sortStats(tchanged)
	
	for k,v in orderedPairs(stats) do
		k = prettyName[k]
		name = v[1]
		val = v[2]+0
		if statIsEnabled(prettyName[name]) then
			if (val > 0) then
				tooltip:AddLine("|cff00ff00+"..val.."|cffffffff "..prettyName[name])
			elseif (val < 0) then
				tooltip:AddLine("|cffff0000"..val.."|cffffffff "..prettyName[name])
			end
		end
	end
	
	if (tooltip:GetName() == "GameTooltip") then
		tooltip:Show()
	end
end

local function hasTwoSlots(loc)
	if (loc == "INVTYPE_TRINKET" or loc == "INVTYPE_FINGER" or loc == "INVTYPE_WEAPON") then
		return true
	else
		return false
	end
end


--[[function getNewStats(link)
	--local _,id,_,_,_,_,_,suffix,uid = strsplit(":",link)
	--return GetItemStats("item:"..id..":0:0:0:0:0:"..suffix..":"..uid)
	return GetItemStats(link)
end]]

local function getCurrentStats(link)
	if (link == nil) then
		return {}
	else
		return GetItemStats(link)
	end
end

--[[
function iStats(id,suffix,uid)
	for k,v in pairs(GetItemStats("item:"..id..":0:0:0:0:0:"..suffix..":"..uid)) do
		print(k,v)
	end
end
function iString(link)
	print(link:gsub("\124","\124\124"))
end
function strstats1(link)
	for k,v in pairs(GetItemStats(link)) do -- returns massive stats
		print(k,v)
	end
end
function strstats2(link)
	local _,id,_,_,_,_,_,suffix,uid = strsplit(":",link)
	
	for k,v in pairs(GetItemStats("item:"..id..":0:0:0:0:0:"..suffix..":"..uid)) do
		print(k,v)
	end
end
]]

local function checkWeaponType(type)
	_,class = UnitClass("player")
	
	if class == "DEATHKNIGHT" then
		if type == "One-Handed Axes"
		or type == "One-Handed Maces"
		or type == "One-Handed Swords"
		or type == "Two-Handed Axes"
		or type == "Two-Handed Maces"
		or type == "Two-Handed Swords"
		or type == "Polearms"
		then return true else return false end
	
	elseif class == "DRUID" then
		if type == "Daggers"
		or type == "Fist Weapons"
		or type == "One-Handed Maces"
		or type == "Polearms"
		or type == "Staves"
		or type == "Two-Handed Maces"
		or type == ""
		or type == ""
		or type == ""
		or type == ""
		or type == ""
		then return true else return false end
		
	elseif class == "HUNTER" then
		if type == "Bows"
		or type == "Crossbows"
		or type == "Daggers"
		or type == "Guns"
		or type == "Fist Weapons"
		or type == "One-Handed Axes"
		or type == "One-Handed Swords"
		or type == "Polearms"
		or type == "Staves"
		or type == "Two-Handed Axes"
		or type == "Two-Handed Swords"
		then return true else return false end
	
	elseif class == "MAGE" then
		if type == "Daggers"
		or type == "Staves"
		or type == "One-Handed Swords"
		or type == "Wands"
		then return true else return false end
	
	elseif class == "PALADIN" then
		if type == "One-Handed Axes"
		or type == "One-Handed Maces"
		or type == "Polearms"
		or type == "One-Handed Swords"
		or type == "Two-Handed Axes"
		or type == "Two-Handed Maces"
		or type == "Two-Handed Swords"
		then return true else return false end
	
	elseif class == "PRIEST" then
		if type == "Daggers"
		or type == "One-Handed Maces"
		or type == "Staves"
		or type == "Wands"
		then return true else return false end
	
	elseif class == "ROGUE" then
		if type == "One-Handed Axes"
		or type == "Daggers"
		or type == "Fist Weapons"
		or type == "One-Handed Maces"
		or type == "One-Handed Swords"
		or type == "Bows"
		or type == "Crossbows"
		or type == "Guns"
		then return true else return false end
	
	elseif class == "SHAMAN" then
		if type == "One-Handed Axes"
		or type == "Daggers"
		or type == "Fist Weapons"
		or type == "One-Handed Maces"
		or type == "Staves"
		or type == "Two-Handed Axes"
		or type == "Two-Handed Maces"
		then return true else return false end
	
	elseif class == "WARLOCK" then
		if type == "Daggers"
		or type == "Staves"
		or type == "One-Handed Swords"
		or type == "Wands"
		then return true else return false end
	
	elseif class == "WARRIOR" then
		if type == "One-Handed Axes"
		or type == "Daggers"
		or type == "Fist Weapons"
		or type == "One-Handed Maces"
		or type == "Polearms"
		or type == "Staves"
		or type == "One-Handed Swords"
		or type == "Two-Handed Axes"
		or type == "Two-Handed Maces"
		or type == "Two-Handed Swords"
		or type == "Bows"
		or type == "Crossbows"
		or type == "Guns"
		then return true else return false end
	
	elseif class == "MONK" then
		if type == "Fist Weapons"
		or type == "One-Handed Axes"
		or type == "One-Handed Maces"
		or type == "One-Handed Swords"
		or type == "Polearms"
		or type == "Staves"
		then return true else return false end
	
	end
end

local function handleTooltip(self, ...)
	tooltip = self
	local name, item = tooltip:GetItem()
	if item then -- If this is an item tooltip
		local _,link,rarity,ilevel,level,type,subtype,_,loc = GetItemInfo(item)
		
		if type ~= "Armor" and type ~= "Weapon" then return end			-- If it's not armor or a weapon, quit
		if rarity-1 < SimpleStats.db.profile.minquality then return end	-- If it's below the current quality threshold, quit
		if type == "Weapon" and (SimpleStats.db.profile.usableweaponsonly == true and checkWeaponType(subtype) == false) then return end	-- Check weapon type
		
		if loc == "INVTYPE_TABARD" or loc == "INVTYPE_BODY" then return end -- Filter out shirts/tabards
		
		if type == "Armor" then											-- If it's not Misc (cloak, neck, etc.) or an enabled armor/weapon type, quit
			if subtype ~= "Miscellaneous" then
				if SimpleStats.db.profile[subtype:lower()] == false then return end
			end
		end
		
		local id = tonumber(strmatch(link,"item:(%d+):"))
		local suffix = tonumber(strmatch(link,"item:%d+:%d+:%d+:%d+:%d+:%d+:(%d+):")) or 0
		local uid = tonumber(strmatch(link,"item:%d+:%d+:%d+:%d+:%d+:%d+:%d+:(-?%d+)")) or 0
		
		local e1link = GetInventoryItemLink("player",invTypes[loc])
		if invTypes[loc] == 11 or invTypes[loc] == 13 or invTypes[loc] == 16 then
			local e2link = GetInventoryItemLink("player",invTypes[loc]+1)
		end
		
		if ((hasTwoSlots(loc) == false and id ~= equipid) or (hasTwoSlots(loc) == true)) then -- filter out items we already have equipped - move up top with rest of if/returns
			
			if (addedLine == false) then addedLine = true; tooltip:AddLine(" ") end
			
			if loc == "INVTYPE_TRINKET" then
				tooltip:AddLine("Trinket 1:")
				newStats = GetItemStats(link)
				curStats = getCurrentStats(e1link)
				printStats(newStats,curStats)
				
				tooltip:AddLine(" ")
				tooltip:AddLine("Trinket 2:")
				curStats = getCurrentStats(e2link)
				printStats(newStats,curStats)
			elseif loc == "INVTYPE_FINGER" then
				tooltip:AddLine("Ring 1:")
				newStats = GetItemStats(link)
				curStats = getCurrentStats(e1link)
				printStats(newStats,curStats)
				
				tooltip:AddLine(" ")
				tooltip:AddLine("Ring 2:")
				curStats = getCurrentStats(e2link)
				printStats(newStats,curStats)
			elseif loc == "INVTYPE_WEAPON" then
				tooltip:AddLine("Weapon 1:")
				newStats = GetItemStats(link)
				curStats = getCurrentStats(e1link)
				printStats(newStats,curStats)
				
				tooltip:AddLine(" ")
				tooltip:AddLine("Weapon 2:")
				curStats = getCurrentStats(e2link)
				printStats(newStats,curStats)
			elseif (loc == "INVTYPE_2HWEAPON" or loc == "INVTYPE_RANGED") and equipid2 then -- Looking at a 2H and two 1H are equipped, so combine their stats
				newStats = GetItemStats(link)
				curStats = getCurrentStats(e1link)
				curStats2 = getCurrentStats(e2link)
				printStats(newStats,curStats,curStats2)
				
			else
				if id ~= equipid then
					newStats = GetItemStats(link)
					curStats = getCurrentStats(e1link)
					printStats(newStats,curStats)
				end
			end
		end
	end
end

local function setupTables()
	invTypes = {}
	invTypes["INVTYPE_HEAD"] = 1
	invTypes["INVTYPE_NECK"] = 2
	invTypes["INVTYPE_SHOULDER"] = 3
	invTypes["INVTYPE_BODY"] = 4
	invTypes["INVTYPE_CHEST"] = 5
	invTypes["INVTYPE_ROBE"] = 5
	invTypes["INVTYPE_WAIST"] = 6
	invTypes["INVTYPE_LEGS"] = 7
	invTypes["INVTYPE_FEET"] = 8
	invTypes["INVTYPE_WRIST"] = 9
	invTypes["INVTYPE_HAND"] = 10
	invTypes["INVTYPE_FINGER"] = 11 -- also 12
	invTypes["INVTYPE_TRINKET"] = 13 -- also 14
	invTypes["INVTYPE_CLOAK"] = 15
	invTypes["INVTYPE_WEAPON"] = 16 -- also 17
	invTypes["INVTYPE_SHIELD"] = 17
	invTypes["INVTYPE_2HWEAPON"] = 16
	invTypes["INVTYPE_WEAPONMAINHAND"] = 16
	invTypes["INVTYPE_WEAPONOFFHAND"] = 17
	invTypes["INVTYPE_HOLDABLE"] = 17
	invTypes["INVTYPE_RANGED"] = 16
	invTypes["INVTYPE_RANGEDRIGHT"] = 16
	invTypes["INVTYPE_TABARD"] = 19
	
	prettyName = {}
	prettyName["RESISTANCE0_NAME"] = "Armor"
	prettyName["ITEM_MOD_DAMAGE_PER_SECOND_SHORT"] = "DPS"
	prettyName["EMPTY_SOCKET_META"] = "Meta Sockets"
	prettyName["EMPTY_SOCKET_PRISMATIC"] = "Prismatic Sockets"
	prettyName["EMPTY_SOCKET_COGWHEEL"] = "Cogwheel Sockets"
	prettyName["EMPTY_SOCKET_RED"] = "Red Sockets"
	prettyName["EMPTY_SOCKET_BLUE"] = "Blue Sockets"
	prettyName["EMPTY_SOCKET_YELLOW"] = "Yellow Sockets"
	prettyName["ITEM_MOD_STAMINA_SHORT"] = "Stamina"
	prettyName["ITEM_MOD_AGILITY_SHORT"] = "Agility"
	prettyName["ITEM_MOD_STRENGTH_SHORT"] = "Strength"
	prettyName["ITEM_MOD_INTELLECT_SHORT"] = "Intellect"
	prettyName["ITEM_MOD_CRIT_RATING_SHORT"] = "Crit"
	prettyName["ITEM_MOD_MASTERY_RATING_SHORT"] = "Mastery"
	prettyName["ITEM_MOD_HIT_RATING_SHORT"] = "Hit"
	prettyName["ITEM_MOD_SPIRIT_SHORT"] = "Spirit"
	prettyName["ITEM_MOD_HASTE_RATING_SHORT"] = "Haste"
	prettyName["ITEM_MOD_EXPERTISE_RATING_SHORT"] = "Expertise"
	prettyName["ITEM_MOD_DODGE_RATING_SHORT"] = "Dodge"
	prettyName["ITEM_MOD_PARRY_RATING_SHORT"] = "Parry"
	prettyName["ITEM_MOD_SPELL_POWER_SHORT"] = "Spellpower"
	prettyName["ITEM_MOD_SPELL_PENETRATION_SHORT"] = "Spell Penetration"
	prettyName["ITEM_MOD_ATTACK_POWER_SHORT"] = "Attack Power"
	prettyName["ITEM_MOD_PVP_POWER_SHORT"] = "PvP Power"
	prettyName["ITEM_MOD_RESILIENCE_RATING_SHORT"] = "PvP Resilience"
	prettyName["ITEM_MOD_ARCANE_RESISTANCE_SHORT"] = "Arcane Resistance"
	prettyName["ITEM_MOD_SHADOW_RESISTANCE_SHORT"] = "Shadow Resistance"
	prettyName["ITEM_MOD_FROST_RESISTANCE_SHORT"] = "Frost Resistance"
	prettyName["ITEM_MOD_FIRE_RESISTANCE_SHORT"] = "Fire Resistance"
	prettyName["ITEM_MOD_NATURE_RESISTANCE_SHORT"] = "Nature Resistance"
	
	order = {}
	order["Armor"] = 1
	order["DPS"] = 2
	order["Meta Sockets"] = 3
	order["Prismatic Sockets"] = 4
	order["Cogwheel Sockets"] = 5
	order["Red Sockets"] = 6
	order["Blue Sockets"] = 7
	order["Yellow Sockets"] = 8
	order["Stamina"] = 9
	order["Agility"] = 10
	order["Strength"] = 11
	order["Intellect"] = 12
	order["Spirit"] = 13
	order["Hit"] = 14
	order["Crit"] = 15
	order["Haste"] = 16
	order["Expertise"] = 17
	order["Dodge"] = 18
	order["Parry"] = 19
	order["Mastery"] = 20
	order["Spellpower"] = 21
	order["Spell Penetration"] = 22
	order["Attack Power"] = 23
	order["PvP Power"] = 24
	order["PvP Resilience"] = 25
	order["Fire Resistance"] = 26
	order["Nature Resistance"] = 27
	order["Frost Resistance"] = 28
	order["Arcane Resistance"] = 29
	order["Shadow Resistance"] = 30
end

local function hideBlizzComparison(self) -- Copied from OldComparison - http://www.wowinterface.com/downloads/fileinfo.php?id=14454
	local old = self.SetHyperlinkCompareItem
	self.SetHyperlinkCompareItem = function(self, link, level, shift, main, ...)
		main = nil
		return old(self, link, level, shift, main, ...)
	end
end

function SimpleStats:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("SimpleStatsDB", defaults, "Default")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SimpleStats", options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SimpleStats","SimpleStats")
	SimpleStats:RegisterChatCommand("ss", "ChatCommand")
	SimpleStats:RegisterChatCommand("simplestats", "ChatCommand")
	
	local options2 = deepcopy(options) -- Required to use Blizzard's addon options tabbing, not Ace's (thanks to InCombatIndicator)
	options2.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(SimpleStats.db)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SimpleStats Profile", options2.args.profile)
	self.profileFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SimpleStats Profile", "Profile", "SimpleStats")
	
	handled = false
	addedLine = false
	setupTables()
	
	GameTooltip:HookScript("OnTooltipCleared",resetTooltip)
	GameTooltip:HookScript("OnTooltipSetItem",handleTooltip)
	ItemRefTooltip:HookScript("OnTooltipSetItem",handleTooltip)

	hideBlizzComparison(ShoppingTooltip1)
	hideBlizzComparison(ShoppingTooltip2)
	hideBlizzComparison(ShoppingTooltip3)
	hideBlizzComparison(ItemRefShoppingTooltip1)
	hideBlizzComparison(ItemRefShoppingTooltip2)
	hideBlizzComparison(ItemRefShoppingTooltip3)
end

--- Misc Lua Functions ---

if not deepcopy then
function deepcopy(object)
	local lookup_table = {}
	local function _copy(object)
		if type(object) ~= "table" then
			return object
		elseif lookup_table[object] then
			return lookup_table[object]
		end
		local new_table = {}
		lookup_table[object] = new_table
		for index, value in pairs(object) do
			new_table[_copy(index)] = _copy(value)
		end
		return setmetatable(new_table, getmetatable(object))
	end
	return _copy(object)
end
end
function __genOrderedIndex(t)
    local orderedIndex = {}
    for key in pairs(t) do table.insert( orderedIndex, key ) end
    table.sort( orderedIndex )
    return orderedIndex
end
function orderedNext(t, state)
    if state == nil then
        t.__orderedIndex = __genOrderedIndex( t )
        key = t.__orderedIndex[1]
        return key, t[key]
    end
    key = nil
    for i = 1,table.getn(t.__orderedIndex) do
        if t.__orderedIndex[i] == state then
            key = t.__orderedIndex[i+1]
        end
    end
    if key then return key, t[key] end
    t.__orderedIndex = nil
    return
end
function orderedPairs(t)
    return orderedNext, t, nil
end